from collections import UserDict
from datetime import datetime, date
import pickle
import re

# class Tags - key words, we add it to notes, as class Phones in home work, max lenght = 20
class Tag():
    def __init__(self, value):
        self.value = value
     

    @property
    def value(self):
        return self.__value

    @value.setter
    def value(self, value: str):
        if len(value) > 20:
            raise ValueError
        self.__value = value

    def __str__(self):
        return f"{self.value}"

    def __repr__(self):
        return f"{self.value}"

    def __eq__(self, other):
       return self.value == other.value

    def __gt__(self, other):
       return self.value > other.value


# class Note. It has required field - title, text_content, optional field - tags
class Note():
    def __init__(self, text_content: str, *tag):
        self.text = text_content
        self.number = None
        self.tags = []
        self.note_date = datetime.today().date()   # datetime.now().timestamp()
        # if tag:
        #     self.tags.append(tag)
        for i in tag:
            self.tags.append(Tag(i))

# add tag to note
    def add_tag(self, tag: Tag):
        if tag not in self.tags:
            return self.tags.append(tag)
        return self.tags

        

# delete tag
    def del_tag(self, tag: Tag):
        for p in self.tags:
            if p.value == tag.value:
                self.tags.remove(p)
                return f'Phone {p.value} delete successful.'
        return f'Phone {tag.value} not found' 

# replace one note to another
    def change_note(self, new_text):
        self.text = new_text
        return "Ok"


# long representation. Note text cats to 50 symbols 
    def __repr__(self):
        if len(self.text) < 50:
            text = self.text
        else: 
           text = self.text[0:50] + '...'
        if not self.tags: 
            tags = 'empty'
        else:
            tags =  self.show_all_tags()     
        #return   "Title: {:<10}| Date: {:<12}| Tags: {:<12}| Text: {:<10}".format(self.title, self.note_date, self.tags, text)
        return   f"Number: {self.number}, Date: {self.note_date}, Tags: {tags}, Text: {text}"

# short representation. Only note text. If text > 50 - print line 50n symb
    def __str__(self):
        i = 0
        pr_str = ''
        pos_space = 0
        if len(self.text) < 50:
            return self.text
        else: 
           
            while i < (len(self.text) - 50):
                pos_space = self.text[i:i + 50].rfind(' ') 
                pr_str = pr_str + self.text[i:i + pos_space] + '\n'
                i += pos_space + 1
            if i < len(self.text):
                pr_str = pr_str +  self.text[i:]
        return pr_str
            
    def  show_all_tags(self):
        return ', '.join(str(p) for p in self.tags)                  

# # searsh one tag      
#     def search_tag(self,  tag: Tag ):
#         if tag in self.tags:
#             return True
#         return False   


# class Notebook == Adressbook, dictionary, where key is unic class counter  
class Notebook(UserDict):
    index = 0

    # add new note with class-count number as key
    def add_note(self, note: Note):
        Notebook.index += 1
        note.number = Notebook.index
        self.data[Notebook.index] = note
        return "note added"

    def search_by_tag(self, tag: Tag):
        note_list = []
        result = ''
        for value in self.data.values():
            if tag in value.tags:
                note_list.append(value)
            else:
                result =  f'This tag is absent'    
        result = note_list    
        return result      
                
    # to do search if we have some words in note text
    def search_by_word_in_note(self, phrase):
        text_coincidence = []
        for value in self.data.values():
            if re.search(phrase, value.text, flags=re.IGNORECASE):
                text_coincidence.append(value)
        return text_coincidence
 

    # to do search if we have date
    def search_by_date(self, day: date):
        date_list = []
        for value in self.data.values():
            if day.date() == (value.note_date):
                date_list.append(value)
           
        return date_list


    #sort list of instances
    def sort_by_date(self):
        object_list = list(self.data.values()) 
        object_list.sort(key=lambda param: param.note_date)  
        return object_list

    @staticmethod 
    def sort_key_tag(note):
        if  note.tags:
            return note.tags[0] 
        else:
            return Tag('')   

    #sort list of instances
    def sort_by_tag(self):
        object_list = list(self.data.values()) 
        object_list.sort(key=Notebook.sort_key_tag)  
        return object_list
        

    # show list of tags
    def show_all_tags(self):
        tags_list = []
        
        for value in self.data.values():
            for item in value.tags:
                if item not in tags_list:
                    tags_list.append(item)
        result = tags_list 
        return result

    # show all dates
    def show_all_dates(self):
        date_list = []
        for value in self.data.values():
            if  value.note_date not in date_list:
                date_list.append(value.note_date)
              
        return  ', '.join(p.strftime('%Y-%m-%d') for p in date_list)  

    #serialize Notebook
    def serializer(self, path_to_book):
        if not self.data:
            return 'Notebook is empty'
        with open(path_to_book, "wb") as fh:
            pickle.dump(self.data, fh)
        return 'Notebook saved'

    #deserialize Notebook
    def deserializer(self, path_to_book):
        keys_list = []
        try:
            with open(path_to_book, "rb") as fh:
                self.data = pickle.load(fh)
                keys_list = list(self.data) 
                Notebook.index = max(keys_list)
                res = f'Saved notebook loaded, number of saved notes is {max(keys_list)} '
        except  FileNotFoundError:
            res = 'No file note_book.txt'

        return res
   

#u = Notebook()
#u.deserializer('notebook.txt')
# print(u.show_all_tags())


from collections import UserDict
from datetime import datetime, timedelta
from itertools import islice
import re


class Field:
    def __init__(self, value):
        self.value = value

    def __str__(self):
        return f"{self.value}"

    def __repr__(self):
        return f"{self.value}"


class Birthday(Field):
    # (рік-місяць-день)
    @property
    def value(self):
        return self.__value

    @value.setter
    def value(self, val: str):
        data = val.split("-")
        if not "".join(data).isdigit():
            raise ValueError
        if int(data[0]) > datetime.now().year or int(data[1]) > 12 or int(data[2]) > 30:
            raise ValueError
        self.__value = val


class Address(Field):
    def __init__(self, value):
        self.value = value


class Email(Field):

    def __repr__(self):
        return self.value

    @property
    def value(self):
        return self.__value

    @value.setter
    def value(self, value: str):
        if value is None:
            self.__value = None
        else:
            result = None
            get_email = re.findall(
                r'\b[a-zA-Z][\w\.]+@[a-zA-Z]+\.[a-zA-Z]{2,}', value)
            if get_email:
                for e in get_email:
                    result = e
            if result is None:
                raise AttributeError(f'Incorrect value provided {value}')
            self.__value = result


class Name(Field):
    def __init__(self, value):
        self.value = value.capitalize()


class Phone(Field):

    @property
    def value(self):
        return self.__value

    @value.setter
    def value(self, val: str):
        if not len(val) == 10 and not len(val) == 13 and not val.lstrip("+").isdigit():
            raise ValueError
        if len(val) == 10:
            val = "+38" + val
        if not val[3] == "0":
            raise ValueError
        self.__value = val


class Record():
    def __init__(self, name: Name, phone: Phone = None, birthday: Birthday = None, email=None, address: Address = None):
        self.name = name
        self.phones = []
        self.birthday = birthday
        self.email = email
        self.address = address
        if phone:
            self.phones.append(phone)

    def __str__(self):
        return f"{self.name} - {', '.join([str(p) for p in self.phones])}"

    def __repr__(self):
        return f"{self.name} - {', '.join([str(p) for p in self.phones])}"

    def add_phone(self, phone: Phone):
        self.phones.append(phone)

    def delete_phone(self, phone: Phone):
        for p in self.phones:
            if p.value == phone.value:
                self.phones.remove(p)
                return f'Phone {p.value} delete successful.'
        return f'Phone {phone.value} not found'

    def days_to_bd(self):
        # (рік-місяць-день)
        if not self.birthday:
            print("Birthday not entered")
        else:
            date1 = self.birthday.value.split("-")
            date = datetime(year=datetime.now().year, month=int(
                date1[1]), day=int(date1[2]))
            data_now = datetime.now()
            dat = date - data_now
            return dat.days

    def add_email(self, email):
        self.email = email

    def add_address(self, address):
        self.address = address

    def add_birthday(self, birthday):
        self.birthday = birthday


class AddressBook(UserDict):
    index = 0

    def add_record(self, rec: Record):
        self[rec.name.value] = rec

    def __str__(self):
        return '\n'.join([str(i) for i in self.values()])

    def iteration(self, step=5):
        while AddressBook.index < len(self):
            yield list(islice(self.items(), AddressBook.index, AddressBook.index+step))
            if AddressBook.index > len(self):
                raise StopIteration()
            AddressBook.index += step

    def birthday_ib_day(self, day):
        days = datetime.now().date() + timedelta(days=day)
        caunt = 0
        for i in self:
            if self[i].birthday:
                date1 = self[i].birthday.value.split("-")
                date = datetime(year=datetime.now().year, month=int(
                    date1[1]), day=int(date1[2])).date()
                if days == date:
                    print(self[i])
                    caunt += 1
        if caunt == 0:
            print(f"In {day} days there is no birthday")


if __name__ == '__main__':
    name = Name('Bill')
    phone = Phone('0666266830')
    rec = Record(name, phone)
    ab = AddressBook()
    ab.add_record(rec)
    name1 = Name('alisa')
    phone1 = Phone('+380662951095')
    rec1 = Record(name1, phone1)
    rec1.add_birthday(Birthday("1997-02-01"))
    ab.add_record(rec1)
    ab.birthday_ib_day(1)



from collections import UserDict
# from Class import *
from Address_book.Class import *
from datetime import datetime
from prompt_toolkit import print_formatted_text, prompt, HTML
from prompt_toolkit.completion import NestedCompleter
import pickle
try:
    with open("Address_book/AddressBook.txt", "rb") as file:
        ab = pickle.load(file)
except:
    ab = AddressBook()
pages = []


def input_error(func):
    def wrapper(*args):
        try:
            return func(*args)
        except IndexError:
            print("Sorry, try again")
        except ValueError:
            print("incorrect input")
        except KeyError:
            print("incorrect Name")
        except AttributeError:
            print("incorrect Email")
    return wrapper


def greetings(*args):
    print("How can I help you?")


@input_error
def add(*argv):
    phone, birthday = None, None
    name = Name(argv[0][0])
    if len(argv[0]) >= 2:
        phone = Phone(argv[0][1])
    if len(argv[0]) >= 3:
        birthday = Birthday(argv[0][2])
    ab.add_record(Record(name, phone, birthday))
    print(f"Contact {name} added")


@input_error
def add_phon(*argv):
    ab[argv[0][0].title()].add_phone(Phone(argv[0][1]))
    print(f"phone {argv[0][1]} added to contact {argv[0][0].title()}")


@input_error
def change(*argv):
    if len(argv[0]) == 3:
        ab[argv[0][0].title()].add_phone(Phone(argv[0][2]))
        ab[argv[0][0].title()].delete_phone(Phone(argv[0][1]))
        print(
            f"In contact {argv[0][0].title()} number {argv[0][1]} is replaced by {argv[0][2]}")
    else:
        print("To change the number, enter the name of the contact, the number to change, the new phone number")


@input_error
def del_phone(*argv):
    ab[argv[0][0].title()].delete_phone(Phone(argv[0][1]))
    print(f"Phone {argv[0][1]} removed from contact {argv[0][0].title()}")


@input_error
def show_all(*argv):
    print(ab)


@input_error
def page(*argv):
    reg = ab.iteration(2)
    for b in reg:
        pages.append(b)
    print(f"page {int(argv[0][0])} of {len(pages)}")
    for i in pages[int(argv[0][0]) - 1]:
        print(i[1])


@input_error
def output_phone(name):
    print(ab[name[0].title()])
    if ab[name[0].title()].birthday:
        print(f"Birthday: {ab[name[0].title()].birthday}")
    if ab[name[0].title()].email:
        print(f"email: {ab[name[0].title()].email}")
    if ab[name[0].title()].address:
        print(f"address: {ab[name[0].title()].address}")


@input_error
def add_birthday(*args):
    ab[args[0][0].title()].birthday = Birthday(args[0][1])
    print(f"Birthday {args[0][1]} added to contact {args[0][0].title()}")


@input_error
def search(val):
    if val[0].isdigit():
        for p in ab.items():
            for x in p[1].phones:
                if not str(x).find(val[0]) == -1:
                    print(p[1])
    else:
        for p in ab.items():
            if not p[0].lower().find(val[0].lower()) == -1:
                print(p[1])


@input_error
def day_birthday(name):
    print(
        f"Before birthday {name[0].title()}, {ab[name[0].title()].days_to_bd()} days")


@input_error
def add_email(*args):
    ab[args[0][0].title()].email = Email(args[0][1])
    print(f"Email {args[0][1]} added to contact {args[0][0].title()}")


@input_error
def add_address(*args):
    ab[args[0][0].title()].address = Address(args[0][1])
    print(f"Address {args[0][1]} added to contact {args[0][0].title()}")


@input_error
def birthday_ib_day(day):
    print(f"In {day[0]} days the birthday of:")
    ab.birthday_ib_day(int(day[0]))


completer = NestedCompleter.from_nested_dict({"hello": None, "add": None, "append": None, "change": None, "phone": None, "show": None,
                                             "del": None, "birthday": None, "day": None, "page": None, "search": None, "email": None, "happy_birthday": None, "exit": None, "home": None})

COMMANDS = {
    greetings: "hello",
    add: "add",
    add_phon: 'append',
    change: "change",
    output_phone: "phone",
    show_all: "show",
    del_phone: "del_phon",
    add_birthday: "birthday",
    day_birthday: "day",
    page: "page",
    search: "search",
    add_email: "email",
    add_address: "home",
    birthday_ib_day: "happy_birthday"
}


def command_parser(u_input: str):
    for comand, key_words in COMMANDS.items():
        if u_input.startswith(key_words):
            return comand, u_input.replace(key_words, "").strip().split(" ")
    return None, None


def address_Book():
    print("Address book open")

    while True:
        u_input = prompt('>>>', completer=completer)
        u_input = u_input.lower()
        if u_input in [".", "good bye", "close", "exit", "/", ""]:
            print("Good bye!")
            break
        comand, data = command_parser(u_input)
        if not comand:
            print("Enter command")
        else:
            comand(data)

    with open("Address_book/AddressBook.txt", "wb") as file:
        pickle.dump(ab, file)


if __name__ == "__main__":
    address_Book()
